// Fake_Keyboard.h
// Send synthetic keypresses to an X Display.
//
// Copyright (C) 2000 Bradley D. LaRonde <brad@ltc.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <vector>
#include <X11/extensions/XTest.h>

static KeySym intlsym[] = {
224, 192, 
225, 193,
226, 194,
227, 195,
228, 196,
229, 197,
230, 198
};

class Fake_Keyboard {
private:
	Display* display;
	int minkeycode, maxkeycode;
	KeySym* mapping;
	int keysyms_per;
	std::vector<int> modifier_stack;
	bool meta_key;

	void find(KeySym sym, int* keycode, bool* shift, bool* control) {
		if(mapping == 0)
			return;

		for(int i=0; i < (maxkeycode - minkeycode) * keysyms_per; i++)
			if(mapping[i] == sym) {
				*keycode = (i / keysyms_per) + minkeycode;
				int modifiers = i % keysyms_per;
				*shift = (modifiers == 1) ? true : false;
				*control = false;
				return;
			}

		// KeySym not found.
		*keycode = 0;
		return;
	}

public:
	Fake_Keyboard(Display* d) : mapping(0) {
		set_display(d);
	}

	~Fake_Keyboard() {
		if(mapping != 0)
			XFree(mapping);
	}

	void set_display(Display* d) {
		display = d;

		if(mapping != 0) {
			XFree(mapping);
			mapping = 0;
		}

		XDisplayKeycodes(display, &minkeycode, &maxkeycode);
		
		
//		XChangeKeyboardMapping(display, minkeycode, 2, intlsym, );  

		mapping = XGetKeyboardMapping(display,
			minkeycode, maxkeycode - minkeycode, &keysyms_per);

		for(int i = 0; i < maxkeycode - minkeycode; i++) {
			//printf("%d: ", minkeycode + i);
			for(int j = 0; j < keysyms_per; j++) {
				//printf("%d(%c) ", mapping[(i * keysyms_per) + j], (char)mapping[(i * keysyms_per) + j]);
			}
			//printf("\n");
		}
	}

	void press(KeySym sym) {
		// Figure keycode and modifiers.
		int keycode;
		bool shift, control;
		find(sym, &keycode, &shift, &control);
		
		// No keycode available for this KeySym, so just return.
		if(keycode == 0)
			return;

		// Figure modifier keycodes.
		int shift_keycode = XKeysymToKeycode(display, XK_Shift_L);
		int control_keycode = XKeysymToKeycode(display, XK_Control_L);
		int meta_keycode = XKeysymToKeycode(display, XK_Meta_L);

		// Engage modifiers.
		if(shift)
			XTestFakeKeyEvent(display, shift_keycode, 1, 0);
		if(control)
			XTestFakeKeyEvent(display, control_keycode, 1, 0);
		if(meta_key == true)
			XTestFakeKeyEvent(display, meta_keycode, 1, 0);

		// Press and release key.
		XTestFakeKeyEvent(display, keycode, 1, 0);
		XTestFakeKeyEvent(display, keycode, 0, 0);

		// Disengage modifiers.
		if(shift)
			XTestFakeKeyEvent(display, shift_keycode, 0, 0);
		if(control)
			XTestFakeKeyEvent(display, control_keycode, 0, 0);
		if(meta_key == true) {
			XTestFakeKeyEvent(display, meta_keycode, 0, 0);
		}
						
		release_modifier();
		release_meta();
	}
	
	void press_modifier(KeySym sym) {
		// Figure keycode and modifiers.
		int keycode;
		bool shift, control;
		find(sym, &keycode, &shift, &control);

		// No keycode available for this KeySym, so just return.
		if(keycode == 0)
			return;

		// Press modifier
		XTestFakeKeyEvent(display, keycode, 1, 0);
		modifier_stack.push_back(keycode);
	}

	void release_modifier(KeySym sym) {
		// Figure keycode and modifiers.
		int keycode;
		bool shift, control;
		find(sym, &keycode, &shift, &control);

		// No keycode available for this KeySym, so just return.
		if(keycode == 0)
			return;

		// Release modifier
		XTestFakeKeyEvent(display, keycode, 0, 0);
	}
	
	void release_modifier() {
		for(std::vector<int>::iterator i = modifier_stack.begin(); i != modifier_stack.end(); i++) {
			// Release modifier
			XTestFakeKeyEvent(display, *i, 0, 0);			
		}
		modifier_stack.clear();
	}

	void press_meta() {
		meta_key = true;
	}
	
	void release_meta() {
		meta_key = false;
	}
};
