// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#include <stdio.h>
#include <math.h>

#include "aqWindow.h"

#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Repeat_Button.H>
#include <FL/x.H>
#include <FL/fl_draw.H>
#include <flpda/Widget_Factory.h>

// images
#include "lower_case.xpm"
#include "upper_case.xpm"
#include "symbols.xpm"
#include "numbers.xpm"
#include "lock.xpm"
#include "ctl.xpm"
#include "alt.xpm"
#include "done.xpm"

typedef enum
{
   SHIFT_CHAR = 1,
   NUMBER_CHAR,
   SYMBOL_CHAR,
   UNSHIFT_CHAR,
   BACKSPACE_CHAR = XK_BackSpace,
   SPACE_CHAR = ' '
} special_chars;

typedef enum 
{
   LOWERCASE_MODE = 0,
   UPPERCASE_MODE,
   NUMBER_MODE,
   SYMBOL_MODE
} entry_modes;

typedef enum
{
   MODE_NORMAL = 0,
   MODE_ONCE,
   MODE_LOCK
} entry_mode_state;

static int char_sets[4][32] = 
{ 
   {'a', 's', 'k', NUMBER_CHAR, BACKSPACE_CHAR, SHIFT_CHAR, 'p', 'f', 'n', 'l', 'x', 'u', 't', 'y', 'j', 'r', 'i', 'd', 'b', SYMBOL_CHAR, SPACE_CHAR, UNSHIFT_CHAR, 'z', 'g', 'o', 'w', 'v', 'c', 'e', 'h', 'q', 'm' },
   {'A', 'S', 'K', NUMBER_CHAR, BACKSPACE_CHAR, SHIFT_CHAR, 'P', 'F', 'N', 'L', 'X', 'U', 'T', 'Y', 'J', 'R', 'I', 'D', 'B', SYMBOL_CHAR, SPACE_CHAR, UNSHIFT_CHAR, 'Z', 'G', 'O', 'W', 'V', 'C', 'E', 'H', 'Q', 'M' },
   {'0', '<', '>', NUMBER_CHAR, BACKSPACE_CHAR, SHIFT_CHAR, '(', ')', ',', '[', ']', '+', '=', '-', '9', '8', '.', '%', '$', SYMBOL_CHAR, SPACE_CHAR, UNSHIFT_CHAR, '7', '6', '5', '*', '/', '4', '3', '#', '2', '1' },
   {'/', '\\','|', NUMBER_CHAR, BACKSPACE_CHAR, SHIFT_CHAR, '(', ')', ',', '[', ']', '"', '\'','&', '!', '?', '.', '-', '_', SYMBOL_CHAR, SPACE_CHAR, UNSHIFT_CHAR, '}', '{', '*', '*', '^', ':', ';', '#', ' ', '@' }
};

static int tap_keys[9] =
{
   0,
   0,
   XK_Alt_L,      // Alt
   XK_Return,     // Ret
   XK_BackSpace,  // Bks
   0,
   XK_Escape,     // Esc
   XK_Tab,        // Tab
   XK_Control_L   // Ctl
};

int aqWindow::handle(int evt)
{
   int wanted = 0;
   int* l = get_current_set();

   if(!(wanted = Fl_Window::handle(evt)))
   {
      switch(evt)
      {
         case FL_PUSH:
         {
            int cx = Fl::event_x();
            int cy = Fl::event_y();

            last_zone_ = which_zone(cx, cy);
            zone_ctr_ = 0;

            wanted = 1;
            break;
         }
         case FL_RELEASE:
         {
            int cx = Fl::event_x();
            int cy = Fl::event_y();

            if(which_zone(cx, cy) == last_zone_ && zone_ctr_ == 0)
            {
               switch(tap_keys[last_zone_])
               {
                  default:
                  {
                     hit_key(tap_keys[last_zone_]);
                     break;
                  }
                  case XK_Alt_L:
                  {
                     meta_key_ = !meta_key_;
                     redraw();
                     break;
                  }
                  case XK_Control_L:
                  {
                     ctrl_key_ = !ctrl_key_;
                     redraw();
                     break;
                  }
               }
            }

            wanted = 1;
            break;
         }
         case FL_DRAG:
         {
            int cx = Fl::event_x();
            int cy = Fl::event_y();
            int z = which_zone(cx, cy);

            // Check if we've gone from one zone to another.
            if(z != last_zone_)
            {
               // Ending a stroke, interpret it.
               if(z == 0)
               {
                  int hit = interpret_stroke();

                  if(hit >= 0)
                  {
                     if(!check_special_char(l[hit]))
                     {
                        hit_key(l[hit]);

                        // Check input modes, see if we need to reset
                        if(input_lock_ == MODE_ONCE)
                        {
                           input_lock_ = MODE_NORMAL;
                           input_mode_ = LOWERCASE_MODE;
                           redraw();
                        }
                     }
                     else
                     {
                        redraw();
                     }
                  }
               }
               else
               {
                  if(zone_ctr_ < MAX_ZONES)
                  {
                     zone_list_[zone_ctr_++] = z;
                  }
               }
            }

            last_zone_ = z;

            wanted = 1;
            break;
         }
      }
   }

   return wanted;
}

void done_cb(Fl_Widget*, void*)
{
   exit(0);
}

aqWindow::aqWindow(int width, int height) :
   Fl_Window(width, height),
   input_mode_(LOWERCASE_MODE),
   input_lock_(MODE_NORMAL),
   meta_key_(0),
   ctrl_key_(0)
{
   int* l = get_current_set();

	// Make a fake keyboard that we can use to send fake keypresses to X.
	fake_keyboard_ = new Fake_Keyboard(fl_display);
   
   Fl_Button* done = new Fl_Button(0, 30, 16, 16, "8");
   Fl_Pixmap* donepix = new Fl_Pixmap(done_xpm);
   donepix->label(done);
   done->callback(done_cb);

   end();

   letter_templates_[LOWERCASE_MODE] = new Fl_Pixmap(lower_case_xpm);
   letter_templates_[UPPERCASE_MODE] = new Fl_Pixmap(upper_case_xpm);
   letter_templates_[NUMBER_MODE] = new Fl_Pixmap(numbers_xpm);
   letter_templates_[SYMBOL_MODE] = new Fl_Pixmap(symbols_xpm);
   lock_image_ = new Fl_Pixmap(lock_xpm);
   meta_image_ = new Fl_Pixmap(alt_xpm);
   ctrl_image_ = new Fl_Pixmap(ctl_xpm);
}

void aqWindow::hit_key(int key)
{
   if(ctrl_key_)
   {
      fake_keyboard_->press_modifier(XK_Control_L);
   }
   if(meta_key_)
   {
      fake_keyboard_->press_modifier(XK_Alt_L);
   }

   fake_keyboard_->press(key);

   if(ctrl_key_)
   {
      fake_keyboard_->release_modifier(XK_Control_L);
      ctrl_key_ = 0;
      redraw();
   }
   if(meta_key_)
   {
      fake_keyboard_->release_modifier(XK_Alt_L);
      meta_key_ = 0;
      redraw();
   }
}

void aqWindow::draw()
{
   int left = w() / 2 - h() / 2;
   int* l = get_current_set();

   Fl_Window::draw();

   letter_templates_[input_mode_]->draw(left, 0);

   if(input_lock_ == MODE_LOCK)
   {
      lock_image_->draw(left + h() / 2 - 4, h() / 2 - 4);
   }

   if(meta_key_)
   {
      meta_image_->draw(left + 42, 48);
   }

   if(ctrl_key_)
   {
      ctrl_image_->draw(left + 7, 48);
   }
}

void aqWindow::draw_char(int ch, int dx, int dy)
{
   int cs, cw, cx, cy;

   dy += fl_descent();

   cs = fl_size() - fl_descent();
   cw = (int)fl_width(' ');
   cx = dx;
   cy = dy - cs;

   switch(ch)
   {
      default:
      {
         char s[] = { ch, 0 };
         fl_draw(s, dx, dy);
         break;
      }
      case BACKSPACE_CHAR:
      {
         fl_loop(cx + cw, cy, 
                    cx + cw, cy + cs, 
                    cx,      cy + cs / 2);
         break;
      }
      case SPACE_CHAR:
      {
         fl_loop(cx,          cy, 
                    cx,          cy + cs, 
                    cx + cw,     cy + cs / 2);
         break;
      }
   }
}

/**
 * Indicate which zone the given coordinate is in.
 */
int aqWindow::which_zone(int zx, int zy)
{
   int s;
   int a;
   int dist;
   int angle;

   // Adjust coords
   zx -= w() / 2;
   zy -= h() / 2;

   dist = (int)sqrt(zx * zx + zy * zy);
   angle = (int)(atan2(zx, zy) * 180.0 / M_PI);

   if(angle < -22.5)
   {
      angle += 360;
   }

   if(dist > h() / 4)
   {
      s = 0;
      for(a = -22; a < 360; a += 45)
      {
         if(angle > (a - 45) && angle <= a)
         {
            break;
         }

         s++;
      }
   }
   else
   {
      s = 0;
   }

   return s;
}

int zone_data[32][3] =
{
   /* 0  */ { 6, 0, 0 },
   /* 1  */ { 6, 5, 0 },
   /* 2  */ { 6, 5, 4 },
   /* 3  */ { 5, 6, 0 },
   /* 4  */ { 5, 0, 0 },
   /* 5  */ { 5, 4, 0 },
   /* 6  */ { 4, 5, 6 },
   /* 7  */ { 4, 5, 0 },
   /* 8  */ { 4, 0, 0 },
   /* 9  */ { 4, 3, 0 },
   /* 10 */ { 4, 3, 2 },
   /* 11 */ { 3, 4, 0 },
   /* 12 */ { 3, 0, 0 },
   /* 13 */ { 3, 2, 0 },
   /* 14 */ { 2, 3, 4 },
   /* 15 */ { 2, 3, 0 },
   /* 16 */ { 2, 0, 0 },
   /* 17 */ { 2, 1, 0 },
   /* 18 */ { 2, 1, 8 },
   /* 19 */ { 1, 2, 0 },
   /* 20 */ { 1, 0, 0 },
   /* 21 */ { 1, 8, 0 },
   /* 22 */ { 8, 1, 2 },
   /* 23 */ { 8, 1, 0 },
   /* 24 */ { 8, 0, 0 },
   /* 25 */ { 8, 7, 0 },
   /* 26 */ { 8, 7, 6 },
   /* 27 */ { 7, 8, 0 },
   /* 28 */ { 7, 0, 0 },
   /* 29 */ { 7, 6, 0 },
   /* 30 */ { 6, 7, 8 },
   /* 31 */ { 6, 7, 0 }
};

/**
 * Interpret the stroke entered and reset the buffer.
 */
int aqWindow::interpret_stroke()
{
   int ret = -1;
   int n;

   // Look for a match in the set of zones.
   for(n = 0; n < 32; n++)
   {
      if(zone_list_[0] == zone_data[n][0] &&
         zone_list_[1] == zone_data[n][1] &&
         zone_list_[2] == zone_data[n][2])
      {
         ret = n;
         break;
      }
   }

   // Reset the zone list
   zone_ctr_ = 0;
   memset(zone_list_, 0, sizeof(zone_list_[0]) * MAX_ZONES);

   return ret;
}

/**
 * Check the input character to see if it should influence our
 * input mode.
 * \return 0 if the char didn't affect mode (so should be output)
 * or nonzero otherwise.
 */
int aqWindow::check_special_char(int ch)
{
   int ret = 0;
   
   switch(ch)
   {
      case SHIFT_CHAR:
      {
         change_mode(UPPERCASE_MODE);
         ret = 1;
         break;
      }
      case NUMBER_CHAR:
      {
         change_mode(NUMBER_MODE);
         ret = 1;
         break;
      }
      case SYMBOL_CHAR:
      {
         change_mode(SYMBOL_MODE);
         ret = 1;
         break;
      }
      case UNSHIFT_CHAR:
      {
         change_mode(LOWERCASE_MODE);
         ret = 1;
         break;
      }
   }

   return ret;
}

/**
 * Change the input mode states, tracking multiple hits
 * against a mode button to lock/unlock the mode state.
 */
void aqWindow::change_mode(int mode)
{
   if(input_mode_ != mode)
   {
      input_mode_ = mode;
      input_lock_ = MODE_ONCE;
   }
   else
   {
      input_lock_++;

      // Rotated through the set of locks, so reset.
      if(input_lock_ > MODE_LOCK)
      {
         input_mode_ = LOWERCASE_MODE;
         input_lock_ = MODE_NORMAL;
      }
   }
}

/**
 * Retrieve the current character set for the current input
 * mode.
 */
int* aqWindow::get_current_set()
{
   return(char_sets[input_mode_]);
}
